'use strict';

const { Atom, Rule, Loader, Plugin, extend } = require('webpack-reconfig');

const fs = require('fs');
const path = require('path');
const webpack = require('webpack');
const PnpWebpackPlugin = require('pnp-webpack-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const InterpolateHtmlPlugin = require('react-dev-utils/InterpolateHtmlPlugin');
const ModuleScopePlugin = require('react-dev-utils/ModuleScopePlugin');
const getCSSModuleLocalIdent = require('react-dev-utils/getCSSModuleLocalIdent');
const VueLoaderPlugin = require('vue-loader/lib/plugin');
const getClientEnvironment = require('../env');
const paths = require('../paths');
const ModuleNotFoundPlugin = require('react-dev-utils/ModuleNotFoundPlugin');

// Webpack uses `publicPath` to determine where the app is being served from.
// It requires a trailing slash, or the file assets will get an incorrect path.
const publicPath = paths.servedPath;
// `publicUrl` is just like `publicPath`, but we will provide it to our app
// as %PUBLIC_URL% in `index.html` and `process.env.PUBLIC_URL` in JavaScript.
// Omit trailing slash as %PUBLIC_URL%/xyz looks better than %PUBLIC_URL%xyz.
const publicUrl = publicPath.slice(0, -1);
// Get environment variables to inject into our app.
const env = getClientEnvironment(publicUrl);

// Check if TypeScript is setup
const useTypeScript = fs.existsSync(paths.appTsConfig);

// style files regexes
const cssRegex = /\.css$/;
const cssModuleRegex = /\.module\.css$/;
const sassRegex = /\.(scss|sass)$/;
const sassModuleRegex = /\.module\.(scss|sass)$/;

// "postcss" loader applies autoprefixer to our CSS.
// "css" loader resolves paths in CSS and adds assets as dependencies.
// "style" loader turns CSS into JS modules that inject <style> tags.
// In production, we use a plugin to extract that CSS to a file, but
// in development "style" loader enables hot editing of CSS.
const baseStyleRule = Rule('css', {
  use: [
    Loader('style', {
      loader: require.resolve('style-loader'),
    }),
    Loader('css', {
      loader: require.resolve('css-loader'),
      options: {
        importLoaders: 1,
      },
    }),
    Loader('postcss', {
      // Options for PostCSS as we reference these options twice
      // Adds vendor prefixing based on your specified browser support in
      // package.json
      loader: require.resolve('postcss-loader'),
      options: {
        // Necessary for external CSS imports to work
        // https://github.com/facebook/create-react-app/issues/2677
        ident: 'postcss',
        plugins: () => [
          require('postcss-flexbugs-fixes'),
          require('postcss-preset-env')({
            autoprefixer: {
              flexbox: 'no-2009',
            },
            stage: 3,
          }),
        ],
      },
    }),
  ],
});

// Opt-in support for SASS (using .scss or .sass extensions).
// Chains the sass-loader with the css-loader and the style-loader
// to immediately apply all styles to the DOM.
// By default we support SASS Modules with the
// extensions .module.scss or .module.sass
const sassStyleRule = extend(baseStyleRule, {
  use: [
    Loader('css', {
      options: {
        importLoaders: 2,
      },
    }),
    Loader('sass', {
      loader: require.resolve('sass-loader'),
      options: {
        includePaths: [paths.appStyle],
      },
    }),
  ],
});

module.exports = {
  output: {
    // This is the URL that app is served from. We use "/" in development.
    publicPath: publicPath,
  },
  optimization: {
    // Automatically split vendor and commons
    // https://twitter.com/wSokra/status/969633336732905474
    // https://medium.com/webpack/webpack-4-code-splitting-chunk-graph-and-the-splitchunks-optimization-be739a861366
    splitChunks: {
      chunks: 'all',
      name: false,
    },
    // Keep the runtime chunk separated to enable long term caching
    // https://twitter.com/wSokra/status/969679223278505985
    runtimeChunk: {
      name: 'runtime',
    },
  },
  resolve: Atom({
    // This allows you to set a fallback for where Webpack should look for modules.
    // We placed these paths second because we want `node_modules` to "win"
    // if there are any conflicts. This matches Node resolution mechanism.
    // https://github.com/facebook/create-react-app/issues/253
    modules: ['node_modules'].concat(
      // It is guaranteed to exist because we tweak it in `env.js`
      process.env.NODE_PATH.split(path.delimiter).filter(Boolean)
    ),
    // These are the reasonable defaults supported by the Node ecosystem.
    // We also include JSX as a common component filename extension to support
    // some tools, although we do not recommend using it, see:
    // https://github.com/facebook/create-react-app/issues/290
    // `web` extension prefixes have been added for better support
    // for React Native Web.
    extensions: paths.moduleFileExtensions
      .map(ext => `.${ext}`)
      .filter(ext => useTypeScript || !ext.includes('ts'))
      .concat('.vue'),
    alias: {
      // Support React Native Web
      // https://www.smashingmagazine.com/2016/08/a-glimpse-into-the-future-with-react-native-for-web/
      'react-native': 'react-native-web',
    },
    plugins: [
      // Adds support for installing with Plug'n'Play, leading to faster installs and adding
      // guards against forgotten dependencies and such.
      PnpWebpackPlugin,
      // Prevents users from importing files from outside of src/ (or node_modules/).
      // This often causes confusion because we only process files within src/ with babel.
      // To fix this, we prevent you from importing files out of src/ -- if you'd like to,
      // please link the files into your node_modules/ and let module-resolution kick in.
      // Make sure your source files are compiled, as they will not be processed in any way.
      new ModuleScopePlugin(paths.appSrc, [paths.appPackageJson]),
    ],
  }),
  resolveLoader: Atom({
    plugins: [
      // Also related to Plug'n'Play, but this time it tells Webpack to load its loaders
      // from the current package.
      PnpWebpackPlugin.moduleLoader(module),
    ],
  }),

  module: {
    strictExportPresence: true,
    rules: [
      // Disable require.ensure as it's not a standard language feature.
      { parser: { requireEnsure: false } },

      // First, run the linter.
      // It's important to do this before Babel processes the JS.
      {
        test: /\.(js|mjs|jsx)$/,
        enforce: 'pre',
        use: [
          {
            options: {
              formatter: require.resolve('react-dev-utils/eslintFormatter'),
              eslintPath: require.resolve('eslint'),
              baseConfig: {
                extends: [require.resolve('@fea/eslint-config-fea-app')],
              },
            },
            loader: require.resolve('eslint-loader'),
          },
        ],
        include: paths.appSrc,
      },
      {
        test: /\.vue$/,
        include: paths.appSrc,
        loader: require.resolve('vue-loader'),
        options: {
          transformToRequire: {
            video: ['src', 'poster'],
            source: 'src',
            img: 'src',
            image: 'xlink:href',
          },
        },
      },
      Rule('oneOf', {
        oneOf: [
          Rule('image', {
            test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
            loader: require.resolve('url-loader'),
            options: {
              limit: 10000,
            },
          }),
          // Process application JS with Babel.
          // The preset includes JSX, Flow, and some ESnext features.
          Rule('js', {
            test: /\.(js|mjs|jsx|ts|tsx)$/,
            include: paths.appSrc,
            loader: require.resolve('babel-loader'),
            options: {
              customize: require.resolve('babel-preset-react-app/webpack-overrides'),
              presets: [require.resolve('babel-preset-react-app')],
              plugins: [
                [
                  require.resolve('babel-plugin-named-asset-import'),
                  {
                    loaderMap: {
                      svg: {
                        ReactComponent: '@svgr/webpack?-svgo,+ref![path]',
                      },
                    },
                  },
                ],
              ],
              // This is a feature of `babel-loader` for webpack (not Babel itself).
              // It enables caching results in ./node_modules/.cache/babel-loader/
              // directory for faster rebuilds.
              cacheDirectory: true,
            },
          }),
          // Process any JS outside of the app with Babel.
          // Unlike the application JS, we only compile the standard ES features.
          Rule('externalJs', {
            test: /\.(js|mjs)$/,
            exclude: /@babel(?:\/|\\{1,2})runtime/,
            loader: require.resolve('babel-loader'),
            options: {
              babelrc: false,
              configFile: false,
              compact: false,
              presets: [
                [require.resolve('babel-preset-react-app/dependencies'), { helpers: true }],
              ],
              cacheDirectory: true,
              // If an error happens in a package, it's possible to be
              // because it was compiled. Thus, we don't want the browser
              // debugger to show the original code. Instead, the code
              // being evaluated would be much more helpful.
              sourceMaps: false,
            },
          }),
          extend(baseStyleRule, {
            test: cssRegex,
            resourceQuery: {
              include: /module/,
              exclude: /lang=(scss|sass)/,
            },
            use: [
              Loader('css', {
                options: {
                  modules: true,
                  getLocalIdent: getCSSModuleLocalIdent,
                },
              }),
            ],
          }).as('vueCssModule'),
          extend(baseStyleRule, {
            test: cssRegex,
            exclude: cssModuleRegex,
          }).as('css'),
          // Adds support for CSS Modules (https://github.com/css-modules/css-modules)
          // using the extension .module.css
          extend(baseStyleRule, {
            test: cssModuleRegex,
            use: [
              Loader('css', {
                options: {
                  modules: true,
                  getLocalIdent: getCSSModuleLocalIdent,
                },
              }),
            ],
          }).as('cssModule'),
          extend(sassStyleRule, {
            test: sassRegex,
            resourceQuery: /module/,
            use: [
              Loader('css', {
                options: {
                  modules: true,
                  getLocalIdent: getCSSModuleLocalIdent,
                },
              }),
            ],
          }).as('vueSassModule'),
          extend(sassStyleRule, {
            test: sassRegex,
            exclude: sassModuleRegex,
          }).as('sass'),
          // Adds support for CSS Modules, but using SASS
          // using the extension .module.scss or .module.sass
          extend(sassStyleRule, {
            test: sassModuleRegex,
            use: [
              Loader('css', {
                options: {
                  modules: true,
                  getLocalIdent: getCSSModuleLocalIdent,
                },
              }),
            ],
          }).as('sassModule'),
          // "file" loader makes sure those assets get served by WebpackDevServer.
          // When you `import` an asset, you get its (virtual) filename.
          // In production, they would get copied to the `build` folder.
          // This loader doesn't use a "test" so it will catch all modules
          // that fall through the other loaders.
          Rule('assets', {
            // Exclude `js` files to keep "css" loader working as it injects
            // its runtime that would otherwise be processed through "file" loader.
            // Also exclude `html` and `json` extensions so they get processed
            // by webpacks internal loaders.
            exclude: [/\.(js|mjs|jsx|ts|tsx)$/, /\.vue$/, /\.html$/, /\.json$/],
            loader: require.resolve('file-loader'),
          }),
        ],
      }),
    ],
  },

  plugins: [
    Atom(new VueLoaderPlugin()),
    // Generates an `index.html` file with the <script> injected.
    Plugin('htmlPlugin', {
      plugin: HtmlWebpackPlugin,
      option: {
        inject: true,
      },
    }),
    // Makes some environment variables available in index.html.
    // The public URL is available as %PUBLIC_URL% in index.html, e.g.:
    // <link rel="shortcut icon" href="%PUBLIC_URL%/favicon.ico">
    // In development, this will be an empty string.
    Atom(new InterpolateHtmlPlugin(HtmlWebpackPlugin, env.raw)),
    // This gives some necessary context to module not found errors, such as
    // the requesting resource.
    Atom(new ModuleNotFoundPlugin(paths.appPath)),
    // Makes some environment variables available to the JS code, for example:
    // if (process.env.NODE_ENV === 'development') { ... }. See `./env.js`.
    Atom(new webpack.DefinePlugin(env.stringified)),
    // Moment.js is an extremely popular library that bundles large locale files
    // by default due to how Webpack interprets its code. This is a practical
    // solution that requires the user to opt into importing specific locales.
    // https://github.com/jmblog/how-to-optimize-momentjs-with-webpack
    // You can remove this if you don't use Moment.js:
    Atom(new webpack.IgnorePlugin(/^\.\/locale$/, /moment$/)),
    // Generate a manifest file which contains a mapping of all asset filenames
    // to their corresponding output file so that tools can pick it up without
    // having to parse `index.html`.
    Plugin('manifestPlugin', {
      plugin: require.resolve('webpack-manifest-plugin'),
      option: {
        publicPath: publicPath,
      },
    }),
    // force entryChunk to use output.filename instead of output.chunkName
    Plugin('chunkNamesPlugin', {
      plugin: require.resolve('@fea/chunk-names-webpack-plugin'),
    }),
  ],

  // Some libraries import Node modules but don't use them in the browser.
  // Tell Webpack to provide empty mocks for them so importing them works.
  node: {
    module: 'empty',
    dgram: 'empty',
    dns: 'mock',
    fs: 'empty',
    net: 'empty',
    tls: 'empty',
    child_process: 'empty',
  },
  // Turn off performance processing because we utilize
  // our own hints via the FileSizeReporter
  performance: false,
};
