// @remove-on-eject-begin
/**
 * Copyright (c) 2015-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
// @remove-on-eject-end
'use strict';

const { Atom, Rule, Loader, Plugin, extend, make } = require('webpack-reconfig');

const path = require('path');
const ModuleNotFoundPlugin = require('@fea/dev-utils/ModuleNotFoundPlugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const OptimizeCSSAssetsPlugin = require('optimize-css-assets-webpack-plugin');
const safePostCssParser = require('postcss-safe-parser');
const paths = require('../../paths');

const isProd = process.env.NODE_ENV === 'production';

// Webpack uses `publicPath` to determine where the app is being served from.
// It requires a trailing slash, or the file assets will get an incorrect path.
const publicPath = paths.servedPath;
// Some apps do not use client-side routing with pushState.
// For these, "homepage" can be set to "." to enable relative asset paths.
const shouldUseRelativeAssetPaths = publicPath === './';
// Source maps are resource heavy and can cause out of memory issue for large source files.
const shouldUseSourceMap = isProd && process.env.GENERATE_SOURCEMAP !== 'false';

const cssRegex = /\.css$/;
const sassRegex = /\.(scss|sass)$/;

const baseStyleRule = Rule('css', {
  use: [
    Loader('style', {
      loader: MiniCssExtractPlugin.loader,
      options: Object.assign(
        {},
        shouldUseRelativeAssetPaths ? { publicPath: '../../' } : undefined
      ),
    }),
    Loader('css', {
      loader: require.resolve('css-loader'),
      options: {
        importLoaders: 1,
        sourceMap: shouldUseSourceMap,
      },
    }),
    Loader('postcss', {
      // Options for PostCSS as we reference these options twice
      // Adds vendor prefixing based on your specified browser support in
      // package.json
      loader: require.resolve('postcss-loader'),
      options: {
        // Necessary for external CSS imports to work
        // https://github.com/facebook/create-react-app/issues/2677
        ident: 'postcss',
        plugins: () => [
          require('postcss-flexbugs-fixes'),
          require('postcss-preset-env')({
            autoprefixer: {
              flexbox: 'no-2009',
            },
            stage: 3,
          }),
        ],
        sourceMap: shouldUseSourceMap,
      },
    }),
  ],
});

const sassStyleRule = extend(baseStyleRule, {
  use: [
    Loader('css', {
      options: {
        importLoaders: 2,
      },
    }),
    Loader('sass', {
      loader: require.resolve('sass-loader'),
      options: {
        includePaths: [paths.appStyle],
        sourceMap: shouldUseSourceMap,
      },
    }),
  ],
});

const getOutput = entryConfig => {
  const scope = entryConfig.scope;
  return {
    filename: `bundle/css/${scope}/${entryConfig.filename}`,
    mediaFilename: `bundle/css/${scope}/media/[name].[hash:8].[ext]`,
  };
};

// This is the production configuration.
// It compiles slowly and is focused on producing a fast and minimal bundle.
// The development configuration is different and lives in a separate file.
module.exports = function(entryConfig) {
  const output = getOutput(entryConfig);

  const config = {
    mode: 'production',
    // Don't attempt to continue if there are any errors.
    bail: true,
    // We generate sourcemaps in production. This is slow but gives good results.
    // You can exclude the *.map files from the build during deployment.
    devtool: shouldUseSourceMap ? 'source-map' : false,
    // In production, we only want to load the app code.
    entry: [entryConfig.entryFile],
    output: {
      path: paths.appPublic,
      publicPath: publicPath,
      devtoolModuleFilenameTemplate: info =>
        path.relative(paths.appSrc, info.absoluteResourcePath).replace(/\\/g, '/'),
    },
    optimization: {
      minimize: isProd,
      minimizer: Atom([
        new OptimizeCSSAssetsPlugin({
          cssProcessorOptions: {
            parser: safePostCssParser,
            map: shouldUseSourceMap
              ? {
                  // `inline: false` forces the sourcemap to be output into a
                  // separate file
                  inline: false,
                  // `annotation: true` appends the sourceMappingURL to the end of
                  // the css file, helping the browser find the sourcemap
                  annotation: true,
                }
              : false,
          },
        }),
      ]),
      splitChunks: {
        cacheGroups: {
          styles: {
            name: 'styles',
            test: /\.css$/,
            chunks: 'all',
            enforce: true,
          },
        },
      },
      runtimeChunk: false,
    },
    module: {
      rules: [
        Rule('oneOf', {
          oneOf: [
            Rule('image', {
              test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
              loader: require.resolve('url-loader'),
              options: {
                limit: 10000,
                name: output.mediaFilename,
              },
            }),
            // "postcss" loader applies autoprefixer to our CSS.
            // "css" loader resolves paths in CSS and adds assets as dependencies.
            // "style" loader turns CSS into JS modules that inject <style> tags.
            // In production, we use a plugin to extract that CSS to a file, but
            // in development "style" loader enables hot editing of CSS.
            // By default we support CSS Modules with the extension .module.css
            extend(baseStyleRule, {
              test: cssRegex,
            }).as('css'),
            extend(sassStyleRule, {
              test: sassRegex,
            }).as('sass'),
          ],
        }),
      ],
    },
    plugins: [
      Atom(new ModuleNotFoundPlugin(paths.appPath)),
      Plugin('miniCssExtractPlugin', {
        plugin: MiniCssExtractPlugin,
        option: {
          chunkFilename: output.filename,
        },
      }),
      Atom({
        apply(compiler) {
          compiler.hooks.emit.tap('NoEmitJsPlugin', compilation => {
            // compilation.assets
            Object.keys(compilation.assets).forEach(asset => {
              if (!asset.endsWith('.css')) {
                delete compilation.assets[asset];
              }
            });
          });
        },
      }),
      // Plugin('chunkNamesPlugin', {
      //   plugin: require.resolve('@fea/chunk-names-webpack-plugin'),
      //   option: {
      //     'mini-css-extract-plugin': {

      //     },
      //   },
      // }),
    ],
  };

  return make(config);
};
