// @remove-on-eject-begin
/**
 * Copyright (c) 2015-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
// @remove-on-eject-end
'use strict';

const { Rule, Plugin, extend, make } = require('webpack-reconfig');

const path = require('path');
const webpack = require('webpack');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const InjectScriptHtmlPlugin = require('@fea/dev-utils/InjectScriptHtmlPlugin');
// @remove-on-eject-begin
const getCacheIdentifier = require('react-dev-utils/getCacheIdentifier');
// @remove-on-eject-end
const paths = require('../paths');
const joinPath = require('./utlis').joinPath;

const getOutput = entryConfig => {
  const scope = entryConfig.scope;
  const mainFilename = joinPath(scope, 'static', 'js/bundle.js');
  return {
    mainFilename,
    filename(chunkData) {
      return chunkData.chunk.name === 'main'
        ? mainFilename
        : joinPath(scope, 'static', 'js/[name].js');
    },
    chunkFilename: joinPath(scope, 'static', 'js/[name].chunk.js'),
    mediaFilename: joinPath(scope, 'static', 'media/[name].[hash:8].[ext]'),
    manifestFilename: joinPath(scope, 'asset-manifest.json'),
    htmlFilename: joinPath(scope, 'index.html'),
  };
};
// This is the development configuration.
// It is focused on developer experience and fast rebuilds.
// The production configuration is different and lives in a separate file.
module.exports = function(entryConfig, dllConfig) {
  const baseConfig = require('./baseConfig');
  const output = getOutput(entryConfig);
  const config = {
    mode: 'development',
    // You may want 'eval' instead if you prefer to see the compiled output in DevTools.
    // See the discussion in https://github.com/facebook/create-react-app/issues/343
    devtool: 'cheap-module-source-map',
    // These are the "entry points" to our application.
    // This means they will be the "root" imports that are included in JS bundle.
    entry: [
      require.resolve('../polyfill.js'),
      // Include an alternative client for WebpackDevServer. A client's job is to
      // connect to WebpackDevServer by a socket and get notified about changes.
      // When you save a file, the client will either apply hot updates (in case
      // of CSS changes), or refresh the page (in case of JS changes). When you
      // make a syntax error, this client will display a syntax error overlay.
      // Note: instead of the default WebpackDevServer client, we use a custom one
      // to bring better experience for Fea users. You can replace
      // the line below with these two lines if you prefer the stock client:
      // require.resolve('webpack-dev-server/client') + '?/',
      // require.resolve('webpack/hot/dev-server'),
      require.resolve('@fea/dev-utils/webpackHotDevClient') + `?filename=/${output.mainFilename}`,
      // Finally, this is your app's code:
      entryConfig.entryPath,
      // We include the app code last so that if there is a runtime error during
      // initialization, it doesn't blow up the WebpackDevServer client, and
      // changing JS code would still trigger a refresh.
    ],
    output: {
      // Add /* filename */ comments to generated require()s in the output.
      pathinfo: true,
      // This does not produce a real file. It's just the virtual path that is
      // served by WebpackDevServer in development. This is the JS bundle
      // containing code from all our entry points, and the Webpack runtime.
      filename: output.filename,
      // There are also additional JS chunk files if you use code splitting.
      chunkFilename: output.chunkFilename,
      // Point sourcemap entries to original disk location (format as URL on Windows)
      devtoolModuleFilenameTemplate: info =>
        path.resolve(info.absoluteResourcePath).replace(/\\/g, '/'),
    },
    module: {
      rules: [
        Rule('oneOf', {
          oneOf: [
            Rule('image', {
              options: {
                name: output.mediaFilename,
              },
            }),
            Rule('js', {
              options: {
                // @remove-on-eject-begin
                // Make sure we have a unique cache identifier, erring on the
                // side of caution.
                // We remove this when the user ejects because the default
                // is sane and uses Babel options. Instead of options, we use
                // the @fea/scripts and babel-preset-react-app versions.
                cacheIdentifier: getCacheIdentifier('development', [
                  'babel-plugin-named-asset-import',
                  'babel-preset-react-app',
                  'react-dev-utils',
                  'fea-scripts',
                ]),
                // @remove-on-eject-end
                // Don't waste time on Gzipping the cache
                cacheCompression: false,
              },
            }),
            Rule('externalJs', {
              options: {
                // Don't waste time on Gzipping the cache
                cacheCompression: false,
                // @remove-on-eject-begin
                cacheIdentifier: getCacheIdentifier('development', [
                  'babel-plugin-named-asset-import',
                  'babel-preset-react-app',
                  'react-dev-utils',
                  'fea-scripts',
                ]),
                // @remove-on-eject-end
              },
            }),
            Rule('assets', {
              // Exclude `js` files to keep "css" loader working as it injects
              // its runtime that would otherwise be processed through "file" loader.
              // Also exclude `html` and `json` extensions so they get processed
              // by webpacks internal loaders.
              options: {
                name: output.mediaFilename,
              },
            }),
          ],
        }),
        // ** STOP ** Are you adding a new loader?
        // Make sure to add the new loader(s) before the "image" loader.
      ],
    },
    plugins: [
      Plugin('htmlPlugin', {
        option: {
          template: entryConfig.htmlTemplatePath,
          filename: output.htmlFilename,
        },
      }),
      // This is necessary to emit hot updates (currently CSS only):
      Plugin('HotModuleReplacementPlugin', {
        plugin: webpack.HotModuleReplacementPlugin,
      }),
      // Watcher doesn't work well if you mistype casing in a path so we use
      // a plugin that prints an error when you attempt to do this.
      // See https://github.com/facebook/create-react-app/issues/240
      Plugin('caseSensitivePathsPlugin', {
        plugin: require.resolve('case-sensitive-paths-webpack-plugin'),
      }),
      // If you require a missing module and then `npm install` it, you still have
      // to restart the development server for Webpack to discover it. This plugin
      // makes the discovery automatic so you don't have to restart.
      // See https://github.com/facebook/create-react-app/issues/186
      Plugin('watchMissingNodeModulesPlugin', {
        plugin: require.resolve('react-dev-utils/WatchMissingNodeModulesPlugin'),
      }),
      Plugin('manifestPlugin', {
        option: {
          fileName: output.manifestFilename,
        },
      }),
      dllConfig && Plugin('InjectScriptHtmlPlugin', {
        plugin: InjectScriptHtmlPlugin,
        options: [HtmlWebpackPlugin, [dllConfig.scriptPath]],
      }),
      dllConfig && Plugin('DllReferencePlugin', {
        plugin: webpack.DllReferencePlugin,
        option: {
          context: paths.appSrc,
          manifest: dllConfig.manifest,
        },
      }),
    ].filter(e => !!e),
  };

  return extend(baseConfig, config);
};
