// @remove-on-eject-begin
/**
 * Copyright (c) 2015-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
// @remove-on-eject-end
'use strict';

const { Atom, Rule, Loader, Plugin, extend, make } = require('webpack-reconfig');

const path = require('path');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const InlineChunkHtmlPlugin = require('react-dev-utils/InlineChunkHtmlPlugin');
const TerserPlugin = require('terser-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const OptimizeCSSAssetsPlugin = require('optimize-css-assets-webpack-plugin');
const safePostCssParser = require('postcss-safe-parser');
// @remove-on-eject-begin
const getCacheIdentifier = require('react-dev-utils/getCacheIdentifier');
// @remove-on-eject-end
const paths = require('../paths');
const joinPath = require('./utlis').joinPath;

// Webpack uses `publicPath` to determine where the app is being served from.
// It requires a trailing slash, or the file assets will get an incorrect path.
const publicPath = paths.servedPath;
// Some apps do not use client-side routing with pushState.
// For these, "homepage" can be set to "." to enable relative asset paths.
const shouldUseRelativeAssetPaths = publicPath === './';
// Source maps are resource heavy and can cause out of memory issue for large source files.
const shouldUseSourceMap = process.env.GENERATE_SOURCEMAP !== 'false';
// Some apps do not need the benefits of saving a web request, so not inlining the chunk
// makes for a smoother build process.
const shouldInlineRuntimeChunk = process.env.INLINE_RUNTIME_CHUNK !== 'false';
// Some apps do not need the hash fot entry assets
const shouldHashEntryFilename = process.env.HASH_ENTRY_FILENAME !== 'false';

// Assert this just to be safe.
// Development builds of React are slow and not intended for production.
if (process.env.NODE_ENV !== 'production') {
  throw new Error('Production builds must have NODE_ENV=production.');
}

const cssStyleLoader = [
  Loader('style', {
    loader: MiniCssExtractPlugin.loader,
    options: Object.assign(
      {},
      shouldUseRelativeAssetPaths ? { publicPath: '../../../' } : undefined
    ),
  }),
  Loader('css', {
    options: {
      sourceMap: shouldUseSourceMap,
    },
  }),
  Loader('postcss', {
    options: {
      sourceMap: shouldUseSourceMap,
    },
  }),
];

const sassStyleLoader = extend(cssStyleLoader, [
  Loader('sass', {
    options: {
      sourceMap: shouldUseSourceMap,
    },
  }),
]);

const getOutput = entryConfig => {
  const scope = entryConfig.scope;
  const mainFilename = joinPath(
    scope,
    'static',
    shouldHashEntryFilename ? 'js/bundle.[contenthash:8].js' : 'js/bundle.js'
  );
  return {
    mainFilename,
    filename(chunkData) {
      return chunkData.chunk.name === 'main'
        ? mainFilename
        : joinPath(
            scope,
            'static',
            shouldHashEntryFilename ? 'js/[name].[contenthash:8].js' : 'js/[name].js'
          );
    },
    chunkFilename: joinPath(scope, 'static', 'js/[name].[contenthash:8].chunk.js'),
    // caution! if you change this, remember to sync change to publicPath in style loader
    cssFilename: joinPath(
      scope,
      'static',
      shouldHashEntryFilename ? 'css/[name].[contenthash:8].css' : 'css/[name].css'
    ),
    cssChunkFilename: joinPath(scope, 'static', 'css/[name].[contenthash:8].chunk.css'),
    mediaFilename: joinPath(scope, 'static', 'media/[name].[hash:8].[ext]'),
    manifestFilename: joinPath(scope, 'asset-manifest.json'),
    htmlFilename: joinPath(scope, 'index.html'),
  };
};

// This is the production configuration.
// It compiles slowly and is focused on producing a fast and minimal bundle.
// The development configuration is different and lives in a separate file.
module.exports = function(entryConfig) {
  const baseConfig = require('./baseConfig');
  const output = getOutput(entryConfig);

  const config = {
    mode: 'production',
    // Don't attempt to continue if there are any errors.
    bail: true,
    // We generate sourcemaps in production. This is slow but gives good results.
    // You can exclude the *.map files from the build during deployment.
    devtool: shouldUseSourceMap ? 'source-map' : false,
    // In production, we only want to load the app code.
    entry: [require.resolve('../polyfill.js'), entryConfig.entryPath],
    output: {
      // The build folder.
      path: paths.appBuild,
      // Generated JS file names (with nested folders).
      // There will be one main bundle, and one file per asynchronous chunk.
      // We don't currently advertise code splitting but Webpack supports it.
      filename: output.filename,
      chunkFilename: output.chunkFilename,
      // Point sourcemap entries to original disk location (format as URL on Windows)
      devtoolModuleFilenameTemplate: info =>
        path.relative(paths.appSrc, info.absoluteResourcePath).replace(/\\/g, '/'),
    },
    optimization: {
      minimizer: Atom([
        new TerserPlugin({
          terserOptions: {
            parse: {
              // we want terser to parse ecma 8 code. However, we don't want it
              // to apply any minfication steps that turns valid ecma 5 code
              // into invalid ecma 5 code. This is why the 'compress' and 'output'
              // sections only apply transformations that are ecma 5 safe
              // https://github.com/facebook/create-react-app/pull/4234
              ecma: 8,
            },
            compress: {
              ecma: 5,
              warnings: false,
              // Disabled because of an issue with Uglify breaking seemingly valid code:
              // https://github.com/facebook/create-react-app/issues/2376
              // Pending further investigation:
              // https://github.com/mishoo/UglifyJS2/issues/2011
              comparisons: false,
              // Disabled because of an issue with Terser breaking valid code:
              // https://github.com/facebook/create-react-app/issues/5250
              // Pending futher investigation:
              // https://github.com/terser-js/terser/issues/120
              inline: 2,
            },
            mangle: {
              safari10: true,
            },
            output: {
              ecma: 5,
              comments: false,
              // Turned on because emoji and regex is not minified properly using default
              // https://github.com/facebook/create-react-app/issues/2488
              ascii_only: true,
            },
          },
          // Use multi-process parallel running to improve the build speed
          // Default number of concurrent runs: os.cpus().length - 1
          parallel: true,
          // Enable file caching
          cache: true,
          sourceMap: shouldUseSourceMap,
        }),
        new OptimizeCSSAssetsPlugin({
          cssProcessorOptions: {
            parser: safePostCssParser,
            map: shouldUseSourceMap
              ? {
                  // `inline: false` forces the sourcemap to be output into a
                  // separate file
                  inline: false,
                  // `annotation: true` appends the sourceMappingURL to the end of
                  // the css file, helping the browser find the sourcemap
                  annotation: true,
                }
              : false,
          },
        }),
      ]),
    },
    module: {
      rules: [
        Rule('oneOf', {
          oneOf: [
            Rule('image', {
              options: {
                name: output.mediaFilename,
              },
            }),
            Rule('js', {
              options: {
                // @remove-on-eject-begin
                // Make sure we have a unique cache identifier, erring on the
                // side of caution.
                // We remove this when the user ejects because the default
                // is sane and uses Babel options. Instead of options, we use
                // the @fea/scripts and babel-preset-react-app versions.
                cacheIdentifier: getCacheIdentifier('production', [
                  'babel-plugin-named-asset-import',
                  'babel-preset-react-app',
                  'react-dev-utils',
                  'fea-scripts',
                ]),
                // @remove-on-eject-end
                // Save disk space when time isn't as important
                cacheCompression: true,
                compact: true,
              },
            }),
            Rule('externalJs', {
              options: {
                // Save disk space when time isn't as important
                cacheCompression: true,
                // @remove-on-eject-begin
                cacheIdentifier: getCacheIdentifier('production', [
                  'babel-plugin-named-asset-import',
                  'babel-preset-react-app',
                  'react-dev-utils',
                  'fea-scripts',
                ]),
                // @remove-on-eject-end
              },
            }),
            Rule('vueCssModule', {
              use: cssStyleLoader,
            }),
            Rule('css', {
              use: cssStyleLoader,
              // Don't consider CSS imports dead code even if the
              // containing package claims to have no side effects.
              // Remove this when webpack adds a warning or an error for this.
              // See https://github.com/webpack/webpack/issues/6571
              sideEffects: true,
            }),
            Rule('cssModule', {
              use: cssStyleLoader,
            }),
            Rule('vueSassModule', {
              use: sassStyleLoader,
            }),
            Rule('sass', {
              use: sassStyleLoader,
              // Don't consider CSS imports dead code even if the
              // containing package claims to have no side effects.
              // Remove this when webpack adds a warning or an error for this.
              // See https://github.com/webpack/webpack/issues/6571
              sideEffects: true,
            }),
            Rule('sassModule', {
              use: sassStyleLoader,
            }),
            Rule('assets', {
              // Exclude `js` files to keep "css" loader working as it injects
              // its runtime that would otherwise be processed through "file" loader.
              // Also exclude `html` and `json` extensions so they get processed
              // by webpacks internal loaders.
              options: {
                name: output.mediaFilename,
              },
            }),
          ],
        }),
        // ** STOP ** Are you adding a new loader?
        // Make sure to add the new loader(s) before the "file" loader.
      ],
    },
    plugins: [
      Plugin('htmlPlugin', {
        option: {
          template: entryConfig.htmlTemplatePath,
          filename: output.htmlFilename,
          minify: {
            removeComments: true,
            collapseWhitespace: true,
            removeRedundantAttributes: true,
            useShortDoctype: true,
            removeEmptyAttributes: true,
            removeStyleLinkTypeAttributes: true,
            keepClosingSlash: true,
            minifyJS: true,
            minifyCSS: true,
            minifyURLs: true,
          },
        },
      }),
      // Inlines the webpack runtime script. This script is too small to warrant
      // a network request.
      shouldInlineRuntimeChunk &&
        Plugin('inlineChunkHtmlPlugin', {
          plugin: InlineChunkHtmlPlugin,
          options: [HtmlWebpackPlugin, [/runtime.*[.]js/]],
        }),
      Plugin('miniCssExtractPlugin', {
        plugin: MiniCssExtractPlugin,
        option: {
          // Must be consistent with option of chunkNamesPlugin.
          // The webpack runtime requires this to properly generate aysnc chunk name.
          chunkFilename: output.cssChunkFilename,
        }
      }),
      Plugin('manifestPlugin', {
        option: {
          fileName: output.manifestFilename,
        },
      }),
      Plugin('chunkNamesPlugin', {
        option: {
          'mini-css-extract-plugin': {
            filename: output.cssFilename,
            chunkFilename: output.cssChunkFilename,
          },
        },
      }),
    ].filter(Boolean),
  };

  return extend(baseConfig, config);
};
