const fs = require('fs-extra');
const chalk = require('chalk');
const path = require('path');
const program = require('commander');
const spawn = require('react-dev-utils/crossSpawn');
const paths = require('../config/paths');

const templateRoot = path.join(paths.appPath, 'page-templates');
const pageRoot = path.join(paths.appSrc, 'pages');

function toAbsolutePath(filepath) {
  return path.resolve(process.cwd(), filepath);
}

function executeNodeScript({ cwd, args }, data, source) {
  return new Promise((resolve, reject) => {
    const child = spawn(process.execPath, [...args, '-e', source, '--', JSON.stringify(data)], {
      cwd,
      stdio: 'inherit',
    });

    child.on('close', code => {
      if (code !== 0) {
        reject({
          command: `node ${args.join(' ')}`,
        });
        return;
      }
      resolve();
    });
  });
}

program
  .usage('[options] <page-name>')
  .option('-t, --template <template-name>', 'select a template', 'default')
  .option(
    '-d, --template-dir <template-dir>',
    'select a template dir',
    toAbsolutePath,
    templateRoot
  )
  .parse(process.argv);

if (program.args.length <= 0) {
  console.error('error: page name is required.');
  program.outputHelp();
  process.exit(-1);
}

async function installTemplate(templateName, templateRoot, pageName, pageRoot) {
  const templatePath = path.join(templateRoot, templateName);
  const pagePath = path.join(pageRoot, pageName);

  // 不存在此模板
  if (!fs.existsSync(templatePath)) {
    console.error(new Error(`template not exists '${templatePath}'`));
    process.exit(-1);
  }

  if (fs.existsSync(pagePath)) {
    console.error(new Error(`page already exists '${pagePath}'`));
    process.exit(-1);
  }

  const originalDirectory = process.cwd();
  const scriptsDirectory = path.join(pagePath, 'scripts');
  try {
    fs.copySync(templatePath, pagePath, {
      overwrite: false,
      errorOnExist: true,
    });

    const initScriptPath = path.join(scriptsDirectory, 'init.js');
    if (fs.existsSync(initScriptPath)) {
      await executeNodeScript(
        {
          cwd: originalDirectory,
          args: [],
        },
        [pagePath, pageName, templateName, originalDirectory],
        `
        var init = require('${initScriptPath}');
        init.apply(null, JSON.parse(process.argv[1]));
      `
      );
      fs.removeSync(scriptsDirectory);
    }
  } catch (reason) {
    if (reason.command) {
      console.log(`  ${chalk.cyan(reason.command)} has failed.`);
    } else {
      console.log(chalk.red('Unexpected error. Please report it as a bug:'));
      console.log(reason);
    }
    process.exit(-1);
  }

  console.log('\n');
  console.log(` start：yarn start ${pageName}`);
  console.log(` build：yarn build ${pageName}`);
}

installTemplate(program.template, program.templateDir, program.args[0], pageRoot);
