const crypto = require('crypto');
const chalk = require('chalk');
const fs = require('fs-extra');
const path = require('path');
const clearConsole = require('react-dev-utils/clearConsole');
const paths = require('../../config/paths');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');

const isInteractive = process.stdout.isTTY;

function checkWebpackDll() {
  const appPackageJson = require(paths.appPackageJson);
  const dependencies = appPackageJson.dllDependencies || {};
  const packages = Object.keys(dependencies).sort();
  if (packages.length <= 0) {
    return Promise.resolve();
  }

  const dependencyId = packages.map(pkg => `${pkg}@${dependencies[pkg]}`).join('');
  const dependencyHash = crypto
    .createHash('md5')
    .update(dependencyId)
    .digest('hex')
    .slice(0, 8);

  const cacheDirectory = paths.appCache;

  const dllFileName = `dll.${dependencyHash}.js`;
  const manifestFileName = `manifest.${dependencyHash}.json`;
  const dllFilePath = path.join(cacheDirectory, dllFileName);
  const dllConfig = {
    scriptPath: dllFileName,
    manifest: path.join(cacheDirectory, manifestFileName),
  };

  if (fs.existsSync(dllFilePath)) {
    return Promise.resolve(dllConfig);
  }

  fs.emptyDirSync(cacheDirectory);
  const createDllConfig = require('../../config/webpack/config.dll');
  const config = createDllConfig({
    entryFiles: packages,
    outputDir: cacheDirectory,
    dllFileName: dllFileName,
    manifestFileName: manifestFileName,
  });
  return build(config).then(() => dllConfig);
}

function build(config) {
  if (isInteractive) {
    clearConsole();
  }
  console.log(chalk.cyan('Creating a pre-bundled dll file to speed up development...\n'));

  const webpack = require('webpack');
  const compiler = webpack(config);
  return new Promise((resolve, reject) => {
    compiler.run((err, stats) => {
      let messages;
      if (err) {
        if (!err.message) {
          return reject(err);
        }
        messages = formatWebpackMessages({
          errors: [err.message],
          warnings: [],
        });
      } else {
        messages = formatWebpackMessages(
          stats.toJson({ all: false, warnings: true, errors: true })
        );
      }
      if (messages.errors.length) {
        // Only keep the first error. Others are often indicative
        // of the same problem, but confuse the reader with noise.
        if (messages.errors.length > 1) {
          messages.errors.length = 1;
        }
        return reject(new Error(messages.errors.join('\n\n')));
      }
      if (messages.warnings.length) {
        return reject(new Error(messages.warnings.join('\n\n')));
      }

      return resolve();
    });
  });
}

module.exports = checkWebpackDll;
